#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import numpy as np
import nengo.utils.numpy as npext

from nengo_bio.common import Excitatory, Inhibitory
from nengo_bio.ensemble import Ensemble
from nengo_bio.neurons import MultiChannelNeuronType

from nengo.builder.operator import Reset
import nengo.dists
import nengo.builder.ensemble
import nengo.builder.signal


built_attrs = nengo.builder.ensemble.built_attrs + ["synapse_types", "tuning"]
class BuiltEnsemble(collections.namedtuple('BuiltEnsemble', built_attrs)):
    pass

@nengo.builder.Builder.register(Ensemble)
def build_ensemble(model, ens):
    # Fetch the random number generator
    rng = np.random.RandomState(model.seeds[ens])

    # Select the synapse types
    if ens._p_exc is None:
        # Mark all neurons as both excitatory and inhibitory if no probabilities
        # are given
        synapse_types_exc = np.ones(ens.n_neurons, dtype=np.bool)
        synapse_types_inh = np.ones(ens.n_neurons, dtype=np.bool)
    else:
        # Otherwise select a random neuron type per neuron
        synapse_types_exc = rng.choice(
            [True, False], ens.n_neurons, p=(ens.p_exc, ens.p_inh))
        synapse_types_inh = ~synapse_types_exc

    # Store the model parameters in the extended BuiltEnsemble named tuple
    synapse_types = {
        Excitatory: synapse_types_exc,
        Inhibitory: synapse_types_inh
    }

    # If this ensemble uses a default neuron type, just call the default
    # ensemble build function.
    if not isinstance(ens.neuron_type, MultiChannelNeuronType):
        nengo.builder.ensemble.build_ensemble(model, ens)
        model.params[ens] = BuiltEnsemble(*model.params[ens], synapse_types, None)
        return

    # Otherwise generate the evaluation points, encoders, gains, biases manually
    eval_points = nengo.builder.ensemble.gen_eval_points(
        ens, ens.eval_points, rng=rng)
    if isinstance(ens.encoders, nengo.dists.Distribution):
        encoders = nengo.dists.get_samples(
            ens.encoders, ens.n_neurons, ens.dimensions, rng=rng)
    else:
        encoders = npext.array(ens.encoders, min_dims=2, dtype=np.float64)
    if ens.normalize_encoders:
        encoders /= npext.norm(encoders, axis=1, keepdims=True)
    gain, bias, max_rates, intercepts = \
        nengo.builder.ensemble.get_gain_bias(ens, rng)
    scaled_encoders = encoders * (gain / ens.radius)[:, np.newaxis]

    # Store the parameters in the BuiltEnsemble instance
    model.params[ens] = BuiltEnsemble(eval_points=eval_points,
                                      encoders=encoders,
                                      intercepts=intercepts,
                                      max_rates=max_rates,
                                      scaled_encoders=scaled_encoders,
                                      gain=gain,
                                      bias=bias,
                                      synapse_types=synapse_types,
                                      tuning=None)

    # Call the "tune" function of the neuron type to give the neuron type
    # the opportunity to find some optimal model parameters
    tuning = ens.neuron_type.tune(model.dt, model, ens)
    model.params[ens] = model.params[ens]._replace(tuning=tuning)

    # Setup dummy input signals causing the original nengo code to work
    n_neurons, n_inputs = ens.n_neurons, ens.neuron_type.n_inputs
    sig = nengo.builder.signal.Signal(
        np.zeros(0), name="{}.neuron_in".format(ens))
    model.sig[ens.neurons]['in'] = sig
    model.sig[ens]['in'] =  None

    # Setup the actual input signals
    for i in range(n_inputs):
        sig = nengo.builder.signal.Signal(
            np.zeros(n_neurons), name="{}.neuron_in".format(ens))
        model.sig[ens.neurons]['in_{}'.format(i)] = sig
        model.add_op(Reset(sig))

    # Output signal
    sig = nengo.builder.signal.Signal(
        np.zeros(n_neurons), name="{}.neuron_out".format(ens))
    model.sig[ens.neurons]['out'] = sig
    model.sig[ens]['out'] = sig

    # This adds the neuron's operator and sets other signals
    model.build(ens.neuron_type, ens.neurons)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from nengo_bio.internal.multi_compartment_lif_sim import make_simulator_class
from nengo_bio.internal.multi_compartment_lif_parameters import (
    SomaticParameters,
    DendriticParameters,
)

_tmp_dir = None  # Directory compiled libraries are stored in
_compiled_library_map = {}  # Map storing the Simulator objects
_supports_cpp = True


def _generate_simulator_cpp_code(f, params_som, params_den, dt=1e-3, ss=10):
    """
    Generates the C++ code representing a specific neuron simulator and stores
    it in the given file descriptor f. This function is used internally by
    "compile_simulator_cpp".
    """

    def _fmt(x):
        res = "{:g}".format(x)
        if (not "." in res):
            s = res.split('e')
            s[0] += '.0'  # Ensure number is C++ double, just '.' would work, but is ugly
            return 'e'.join(s)
        return res

    # Code below compares the result of _fmt to the string "0.0"
    assert _fmt(0) == "0.0" and _fmt(0.0) == "0.0"

    def _generate_b_vec_entry(i, B, b_const):
        res = _fmt(b_const[i])
        if res == "0.0":
            res = ""
        for k in range(B.shape[1]):
            if B[i, k] != 0.0:
                if len(res) > 0:
                    res += " + "
                res += "({} * x[{}])".format(_fmt(B[i, k]), k)
        if res == "":
            return "0.0"
        return res

    def _generate_A_mat_entry(i, j, A, a_const, C):
        # Lookup off-diagonal entries in the C matrix
        if i != j:
            return _fmt(C[i, j])

        # Assemble diagonal entries according to the Aconst and A matrix
        return _generate_b_vec_entry(i, A, a_const)

    # Handy aliases for params_som and params_den
    pS, pD = params_som, params_den

    # Compute v_min, v_max
    v_min, v_max = params_den.vEq_extreme(params_som)

    f.write(
        """// This code was automatically generated by multi_compartment_lif_cpp.py

#include "multi_compartment_lif.hpp"

namespace { // Do not export the following symbols
struct Parameters {
    static constexpr double tau_ref = """ + _fmt(pS.tau_ref) + """;
    static constexpr double tau_spike = """ + _fmt(pS.tau_spike) + """;
    static constexpr double v_th = """ + _fmt(pS.v_th) + """;
    static constexpr double v_reset = """ + _fmt(pS.v_reset) + """;
    static constexpr double v_spike = """ + _fmt(pS.v_spike) + """;
    static constexpr size_t n_comp = """ + str(pD.n_comp) + """;
    static constexpr size_t n_inputs = """ + str(pD.n_inputs) + """;
    static constexpr double dt = """ + _fmt(dt) + """;
    static constexpr size_t ss = """ + str(ss) + """;

    using MatA = Matrix<double, n_comp, n_comp>;
    using VecASom = Matrix<double, 1, n_comp>;
    using VecB = Matrix<double, n_comp, 1>;
    using VecX = Matrix<double, n_inputs, 1>;
    using VecV = Matrix<double, n_comp, 1>;

    static const VecV v_min;
    static const VecV v_max;

    static MatA calc_A(const VecX &x) {
        (void)x; // x may be unused, suppress warnings
        return (MatA() <<\n""")
    for i in range(pD.n_comp):
        for j in range(pD.n_comp):
            f.write("            {}".format(
                _generate_A_mat_entry(i, j, pD.A, pD.a_const, pD.C)))
            if i != pD.n_comp - 1 or j != pD.n_comp - 1:
                f.write(",")
            f.write("\n")
    f.write("""        ).finished();
    }

    static VecB calc_b(const VecX &x) {
        (void)x; // x may be unused, suppress warnings
        return (VecB() <<\n""")
    for i in range(pD.n_comp):
        f.write("            {}".format(
            _generate_b_vec_entry(i, pD.B, pD.b_const)))
        if i != pD.n_comp - 1:
            f.write(",")
        f.write("\n")
    f.write("""        ).finished();
    }
};

const Parameters::VecV Parameters::v_min{""" + ", ".join(map(_fmt, v_min)) + """};
const Parameters::VecV Parameters::v_max{""" + ", ".join(map(_fmt, v_max)) + """};
}

extern "C" { // Exported C API
void run_step_from_memory(uint32_t n_neurons, double *state, double *out""")
    for i in range(pD.n_inputs):
        f.write(", double *x{}".format(i))
    f.write(""") {
    const double *xs[""" + str(pD.n_inputs) + """] = {""")
    f.write(", ".join("x{}".format(i) for i in range(pD.n_inputs)))
    f.write("""};
    Simulator<Parameters>::run_step_from_memory(n_neurons, state, out, xs);
}

void run_single_with_constant_input(uint32_t n_samples, double *state, double *out, double *xs) {
    Simulator<Parameters>::run_single_with_constant_input(n_samples, state, out, xs);
}

void run_single_with_poisson_sources(uint32_t n_samples, double *state, double *out, PoissonSource *sources) {
    Simulator<Parameters>::run_single_with_poisson_sources(n_samples, state, out, sources);
}

void run_single_with_gaussian_sources(uint32_t n_samples, double *state, double *out, GaussianSource *sources) {
    Simulator<Parameters>::run_single_with_gaussian_sources(n_samples, state, out, sources);
}

};
""")


def _compile_cpp_library(code, tar, debug=False):
    import os
    import subprocess
    import posix_ipc

    # Create a semaphore for the given target file. Only one process should try
    # to compile code at a time
    with posix_ipc.Semaphore(
            os.path.basename(tar), posix_ipc.O_CREAT, initial_value=1) as lock:
        # Someone else compiled the code for us, return!
        if os.path.exists(tar):
            return

        # Fetch the module directory which is used as include directory
        curdir = os.path.dirname(os.path.abspath(__file__))

        # Compile the code
        args = list(
            filter(
                lambda x: bool(x),
                [
                    'g++',
                    '-std=c++11',
                    '-march=native',  # Compile for this particular computer
                    '-g' if debug else '-DNDEBUG',  # Toggle debug code
                    '-O1' if debug else '-O3',  # Toggle optimisation
                    '-Wall',
                    '-Wextra',
                    '-Wno-deprecated-copy',
                    '-fno-exceptions',  # This code does not use exceptions
                    '-fno-rtti',  # We don't need no runtime-type information
                    '-ffast-math',  # Be sloppy, we don't have non-finite math
                    '-I' + os.path.expanduser('~/.local/include/eigen3'),
                    '-I/usr/include/eigen3',
                    '-I' + curdir,
                    '-x',  # Specify the language since we're compiling from stdin
                    'c++',
                    '-fPIC',
                    '-shared',
                    None if debug else '-s',
                    '-Wl,--as-needed,-soname,' + tar,
                    '-lm',
                    '-o',
                    tar
                ]))

        # Write the source code to a temporary file when debugging
        if debug:
            src = tar + '.cpp'
            with open(src, 'wb') as f:
                f.write(code)
            args.append(src)
        else:
            args.append('-')

        # Call the compiler
        process = subprocess.Popen(
            args,
            stdin=None if debug else subprocess.PIPE,
            stderr=subprocess.PIPE)
        _, stderr = process.communicate(None if debug else code)
        stderr = str(stderr, 'utf-8')
        if process.returncode != 0:
            raise Exception("Error while compiling the C++ code:\n" + stderr)
        elif len(stderr) > 0:
            print(stderr)


def _create_tmp_dir():
    import os
    import tempfile

    global _tmp_dir, _compiled_library_map

    # Create a user-accessible temporary directory
    if _tmp_dir is None or not os.path.exists(_tmp_dir):
        _compiled_library_map = {}
        _tmp_dir = os.path.expanduser('~/.cache/multi_compartment_lif/')
        os.makedirs(_tmp_dir, exist_ok=True, mode=0O755)

    return _tmp_dir


def _file_hash(filename, hasher):
    """
    Feeds the content of the file with the specified file into the hasher.
    """

    with open(filename, 'rb') as f:
        while True:
            data = f.read(65536)
            if not data:
                break
            hasher.update(data)


def _compute_code_hash(code):
    import hashlib
    import os

    # Find the header file. Changing this file should also result in a new
    # library being compiled
    header_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "multi_compartment_lif.hpp")

    hasher = hashlib.sha1()
    _file_hash(header_file, hasher)
    _file_hash(__file__, hasher)
    hasher.update(str(os.getuid()).encode('utf-8'))
    hasher.update(code)
    return hasher.hexdigest()


def compile_simulator_cpp(params_som, params_den, dt=1e-3, ss=10):
    """
    Generates and compiles the C++ simulator library. Returns a callable that
    can be used to  To compiled dynamic library will be stored in a temporary
    directory accessible only to the user running this Python process.
    """
    from ctypes import cdll, POINTER, cast, c_double, c_uint32, c_void_p
    import io
    import numpy as np
    import os

    # Generate the C++ code
    f = io.StringIO()
    _generate_simulator_cpp_code(f, params_som, params_den, dt, ss)

    # Compute the hash, generate the lookup table key -- if this neuron model
    # has already been compiled, return the corresponding compiled code
    code = f.getvalue().encode('utf-8')
    sha_hash = _compute_code_hash(code)
    key = "multi_comp_lif_" + sha_hash[:8]
    if not key in _compiled_library_map:
        # Compile the C++ code to a library
        libpath = os.path.join(_create_tmp_dir(), key + '.so')
        if not os.path.exists(libpath):
            _compile_cpp_library(code, libpath)

        # Load the C library
        lib = cdll.LoadLibrary(libpath)
        c_run_step_from_memory = lib.run_step_from_memory
        c_run_single_with_constant_input = lib.run_single_with_constant_input
        c_run_single_with_poisson_sources = lib.run_single_with_poisson_sources
        c_run_single_with_gaussian_source = lib.run_single_with_gaussian_sources

        c_double_p = POINTER(c_double)

        class CppImpl:
            @staticmethod
            def run_step_from_memory(self, out, *xs):
                pstate = self.state.ctypes.data_as(c_double_p)
                pout = out.ctypes.data_as(c_double_p)
                pxs = [x.ctypes.data_as(c_double_p) for x in xs]
                c_run_step_from_memory(
                    c_uint32(self.n_neurons), pstate, pout, *pxs)

            @staticmethod
            def run_single_with_constant_input(self, out, xs):
                pstate = self.state.ctypes.data_as(c_double_p)
                pout = out.ctypes.data_as(c_double_p)
                pxs = xs.ctypes.data_as(c_double_p)
                c_run_single_with_constant_input(
                    c_uint32(out.shape[0]), pstate, pout, pxs)

            @staticmethod
            def run_single_with_poisson_sources(self, out, sources):
                pstate = self.state.ctypes.data_as(c_double_p)
                pout = out.ctypes.data_as(c_double_p)
                psources = cast(sources, c_void_p)
                c_run_single_with_poisson_sources(out.size, pstate, pout,
                                                  psources)

            @staticmethod
            def run_single_with_gaussian_sources(self, out, sources):
                pstate = self.state.ctypes.data_as(c_double_p)
                pout = out.ctypes.data_as(c_double_p)
                psources = cast(sources, c_void_p)
                c_run_single_with_gaussian_source(out.size, pstate, pout,
                                                  psources)

        # Register the above function for the specific neuron type and return it
        _compiled_library_map[key] = make_simulator_class(
            CppImpl, params_som, params_den, dt, ss)
    return _compiled_library_map[key]


def supports_cpp():
    global _supports_cpp
    if _supports_cpp is None:
        import os
        import warnings

        # Create a user-accessible temporary directory
        tar = os.path.join(_create_tmp_dir(), 'test.so')

        # Try to compile a minimal library which just includes eigen
        try:
            compile_cpp_library(
                b"\"multi_compartment_lif.hpp\"\n#include <Eigen/Dense>\n#include <Eigen/Eigenvalues>",
                tar)
            _supports_cpp = True
        except:
            _supports_cpp = False
            warnings.warn(
                "Cannot activate C++ support for multi-compartment LIF "
                "neurons. The simulator will fallback to a Python "
                "implementation. This will be slow. Please make sure you have "
                "a recent version of g++ and Eigen3 installed.",
                RuntimeWarning)

        # Remove the target file
        if os.path.exists(tar):
            os.remove(tar)

    return _supports_cpp





#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

SAMPLE_CACHE = {}

def sample_rates(dt,
                 max_rates,
                 min_max_rate,
                 max_max_rate,
                 run_single_sim,
                 estimate_input_range,
                 filter_input,
                 params_hash,
                 n_samples=1000,
                 rate_discretization=20,
                 T=10.0):
    def halton(i, b):
        f = 1
        r = 0
        while i > 0:
            f = f / b
            r = r + f * (i % b)
            i = i // b
        return r

    def discretize_rate(a):
        return int(np.ceil(a / rate_discretization) * rate_discretization)

    n_inputs = 2  # TODO

    # Create the sample cache map
    if not params_hash in SAMPLE_CACHE:
        SAMPLE_CACHE[params_hash] = {}
    cache = SAMPLE_CACHE[params_hash]

    # Discretize the minimum and maximum rates
    min_max_rate = discretize_rate(min_max_rate)
    max_max_rate = discretize_rate(max_max_rate)
    rate_range = range(max_max_rate, min_max_rate - 1, -rate_discretization)

    # Find the inputs corresponding to the threshold currents
    in_E0, _ = estimate_input_range(1)
    in_E0 = in_E0 * 0.8 # add some safety margin

    # Generate samples for each possible discretization step. Note that
    # this is completely deterministic as long as max_max_rate is the same.
    rate_map = {}
    for i, rate in enumerate(rate_range):
        # Generate a new entry in the rate_map
        rate_map[rate] = {"neurons": set(), "samples": set(), "result": None}

        # Estimate the search space
        in_E, in_I = estimate_input_range(rate)

        # Select as many points as possible from the previous set
        samples = rate_map[rate]["samples"]
        if i > 0:
            for sample in rate_map[rate + rate_discretization]["samples"]:
                if sample[0] < in_E and sample[1] < in_I:
                    samples.add(sample)

        # Fill the set of sample points up to N_SAMPLES samples
        k = int(rate / rate_discretization * n_samples)
        i = len(samples)
        while len(samples) < n_samples:
            sample = (
                in_E0 + halton(k + i, 2) * (in_E - in_E0),
                        halton(k + i, 3) *  in_I)
            i += 1
            if filter_input(*sample):
                samples.add(sample)

    # Iterate over the actual maximum rates and associate each neuron with a
    # slot in the rate map.
    for i, rate in enumerate(max_rates):
        rate = discretize_rate(rate)
        rate_map[rate]["neurons"].add(i)

    # Remove entries that do not have any neurons associated with them
    for key in [key for key in rate_map if len(rate_map[key]["neurons"]) == 0]:
        del rate_map[key]

    # Fetch all samples for which we need to run a simulation
    samples = set()
    for entry in rate_map.values():
        samples = samples.union(entry["samples"])

    # Compute the average rate for each sample by running a single-neuron
    # simulation
    # TODO: Use threading
    samples = list(samples)
    rates = np.empty(len(samples))
    ts = np.arange(0, T, dt)
    for i, sample in enumerate(samples):
        if sample in cache:
            rates[i] = SAMPLE_CACHE[params_hash][sample]
            continue
        out = np.zeros_like(ts)
        run_single_sim(i, out, *sample)
        times = ts[out > 0]
        if (len(times) > 1):
            rates[i] = 1. / np.mean((times[1:] - times[:-1]))
        else:
            rates[i] = 0.0
        cache[sample] = rates[i]

    # Write the measured rates back to the rate_map
    for rate, entry in rate_map.items():
        entry_result = np.empty((len(entry["samples"]), n_inputs + 1))
        j = 0
        for i, sample in enumerate(samples):
            if sample in entry["samples"]:
                entry_result[j, :n_inputs] = sample
                entry_result[j, n_inputs:] = rates[i]
                j += 1
        entry["result"] = entry_result

    return rate_map

def fit_model_weights_two_comp(Js, samples, valid):
    from .qp_solver import solve_linearly_constrained_quadratic_loss

    # Assemble the quadratic loss function
    scale_cur = 1e9
    scale_cond = 1e6
    n_samples = samples.shape[1]
    n_valid = int(np.sum(valid))

    n_cstr = n_valid
    n_vars = 5
    C, d = np.zeros((n_cstr, n_vars)), np.zeros(n_cstr)
    C[:n_valid, 0] = 1
    C[:n_valid, 1] = samples[1, valid] * scale_cond
    C[:n_valid, 2] = -Js[valid] * scale_cur
    C[:n_valid, 3] = -Js[valid] * samples[0, valid] * scale_cur * scale_cond
    C[:n_valid, 4] = -Js[valid] * samples[1, valid] * scale_cur * scale_cond
    d[:n_valid] = -samples[0, valid] * scale_cond

    # Make sure a0, a1, a2 are greater than zero
    G, h = np.zeros((3, n_vars)), np.zeros(3)
    G[0, 2] = -1
    G[1, 3] = -1
    G[2, 4] = -1

    # Solve for model weights
    ws = solve_linearly_constrained_quadratic_loss(C=C, d=d, G=G, h=h, tol=1e-12)[:, 0]

    # Rescale the output
    ws = np.array((
        ws[0] / scale_cond,
        1.0,
        ws[1],
        ws[2] * scale_cur / scale_cond,
        ws[3] * scale_cur,
        ws[4] * scale_cur,
    ))

    return ws


def tune_two_comp_model_weights(dt, max_rates, min_max_rate, max_max_rate,
                                run_single_sim, estimate_input_range,
                                filter_input, lif_rate_inv, params_hash):
    # Generate sample points and measure the
    rate_map = sample_rates(dt, max_rates, min_max_rate, max_max_rate,
                            run_single_sim, estimate_input_range, filter_input, params_hash)

    # Compute the per-neuron model weights
    n_neurons = len(max_rates)
    ws = np.zeros((n_neurons, 6), dtype=np.float64)

    for rate, entry in rate_map.items():
        # Solve for model weights
        result = entry["result"]
        samples, rates = result[:, 0:2], result[:, 2]
        Js = lif_rate_inv(rates)
        valid = rates > 10 # Ignore points with a small rate
        w = fit_model_weights_two_comp(Js, samples.T, valid)

        # Store per-neuron weights in the corresponding slot
        for i in entry["neurons"]:
            ws[i] = w

#        def H(w, gE, gI):
#            return (w[0] + w[1] * gE + w[2] * gI) / (w[3] + w[4] * gE + w[5] * gI)

#        import matplotlib.pyplot as plt
#        import matplotlib
#        matplotlib.use('Agg')

#        gE, gI = samples.T
#        gEs = np.linspace(np.min(gE), np.max(gE))
#        gIs = np.linspace(np.min(gI), np.max(gI))
#        gEss, gIss = np.meshgrid(gEs, gIs)
#        Js_model = H(w, gEss, gIss)

#        fig, ax = plt.subplots()
#        sc = ax.scatter(gE * 1e9, gI * 1e9, c=rates)
#        ax.contour(gEs * 1e9, gIs * 1e9, Js_model, levels=np.linspace(0, np.max(Js), 10), colors=['white'])
#        ax.contour(gEs * 1e9, gIs * 1e9, Js_model, levels=np.linspace(0, np.max(Js), 10), colors=['k'], linestyles='--')
#        ax.set_xlim(0, np.max(gE) * 1e9)
#        ax.set_ylim(0, np.max(gI) * 1e9)
#        ax.set_xlabel("Excitatory input")
#        ax.set_ylabel("Inhibitory input")
#        plt.colorbar(sc)

#        fig.savefig('{}_{}.png'.format(params_hash, rate), dpi=300)

    return ws




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

class EnvGuard:
    """
    Class used to temporarily set some environment variables to a certain value.
    In particular, this is used to set the variable OMP_NUM_THREADS to "1" for
    each of the subprocesses using cvxopt.
    """

    def __init__(self, env):
        self.env = env
        self.old_env_stack = []

    def __enter__(self):
        # Create a backup of the environment variables and write the desired
        # value.
        old_env = {}
        for key, value in self.env.items():
            if key in os.environ:
                old_env[key] = os.environ[key]
            else:
                old_env[key] = None
            os.environ[key] = str(value)

        # Push the old_env object onto the stack of old_env objects.
        self.old_env_stack.append(old_env)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Get the environment variable backup from the stack.
        old_env = self.old_env_stack.pop()

        # Either delete environment variables that were not originally present
        # or reset them to their original value.
        for key, value in old_env.items():
            if not value is None:
                os.environ[key] = value
            else:
                del os.environ[key]

        return False



#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import nengo

def set_defaults():
    """
    Sets the default parameters for ensembles to more biologically realisitc
    defaults. In particular, makes sure that the maximum firing rate is between
    50 and 100 Hz and restricts the intercepts to values between -0.9 and 0.9.
    This prevents target currents from being too large.
    """
    nengo.Ensemble.max_rates.default = nengo.dists.Uniform(50, 100)
    nengo.Ensemble.intercepts.default = nengo.dists.Uniform(-0.9, 0.9)



#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import warnings
import multiprocessing
import os

import cvxopt
import numpy as np
import scipy.optimize

from .env_guard import EnvGuard

DEFAULT_TOL = 1e-6
DEFAULT_REG = 1e-1

class CvxoptParamGuard:
    """
    Class used to set relevant cvxopt parameters and to reset them once
    processing has finished or an exception occurs.
    """

    def __init__(self, tol=1e-24, disp=False):
        self.options = {
            "abstol": tol,
            "feastol": tol,
            "reltol": 10 * tol,
            "show_progress": disp
        }

    def __enter__(self):
        # Set the given options, backup old options
        for key, value in self.options.items():
            if key in cvxopt.solvers.options:
                self.options[key] = cvxopt.solvers.options[key]
            cvxopt.solvers.options[key] = value
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore the old cvxopt options
        for key, value in self.options.items():
            cvxopt.solvers.options[key] = value
        return self

def _solve_qp(Pqp,
              qqp,
              Gqp=None,
              hqp=None,
              Aqp=None,
              bqp=None,
              tol=1e-12,
              disp=True):
    """
    Solves the given quadtratic programing problem

    min    x^T P x + q^T x
    s.t.   Gx <= h
           Ax  = b

    """

    # Solve the QP problem
    with CvxoptParamGuard(tol=tol, disp=disp) as guard:
        res = cvxopt.solvers.qp(
            P=cvxopt.matrix(Pqp.astype(np.double)),
            q=cvxopt.matrix(qqp.astype(np.double)),
            G=None if Gqp is None else cvxopt.matrix(Gqp.astype(np.double)),
            h=None if hqp is None else cvxopt.matrix(hqp.astype(np.double)),
            A=None if Aqp is None else cvxopt.matrix(Aqp.astype(np.double)),
            b=None if bqp is None else cvxopt.matrix(bqp.astype(np.double)))

        return np.array(res["x"])


def _check_basis_pursuit_params(C, d, A, b, G, h):
    """
    Used internally to make sure that the given arguments have the right
    dimensionality.
    """

    # Replace zero-sized matrices with None
    C = None if (not (C is None)) and C.size == 0 else C
    d = None if (not (d is None)) and d.size == 0 else d
    A = None if (not (A is None)) and A.size == 0 else A
    b = None if (not (b is None)) and b.size == 0 else b
    G = None if (not (G is None)) and G.size == 0 else G
    h = None if (not (h is None)) and h.size == 0 else h

    # Make sure either both of the variables in the pairs G, h and A, b are
    # "None" or both of them are not
    assert not (C is None) and not (d is None), "C, d must not be None"
    assert (G is None) == (h is None), "Both G and h must be None"
    assert (A is None) == (b is None), "Both A and b must be None"

    # Make sure d, b, h are vectors
    assert (d is None) or (d.size == d.shape[0]), "d must be a vector"
    assert (b is None) or (b.size == b.shape[0]), "b must be a vector"
    assert (h is None) or (h.size == h.shape[0]), "h must be a vector"

    # Make sure A, C use the same number of variables
    if not A is None:
        assert A.shape[1] == C.shape[1], \
                "Second dimension of A, C, and G (number of variables in " + \
                "the system) must be the same"

    # Make sure C, d and G, h have the same number of constraints
    assert d.shape[0] == C.shape[0], \
            "First dimension of C, d (number of equality constraints) " + \
            "must be the same"
    if not G is None:
        assert G.shape[0] == h.shape[0], \
                "First dimension of G, h (number of inequality " + \
                " constraints) must be the same"

        # Make sure G and C have the same number of variables
        assert G.shape[1] == C.shape[1], \
                "Second dimension of A, C, and G (number of variables in " + \
                "the system) must be the same"

    return C, d, A, b, G, h


def solve_linearly_constrained_quadratic_loss(C,
                                              d,
                                              A=None,
                                              b=None,
                                              G=None,
                                              h=None,
                                              tol=DEFAULT_TOL,
                                              disp=False):
    """
    Solves a problem similar to the basis pursuit problem, but using the L2
    instead of the L1 norm, thus turning the problem into a quadratic program.

    min    || Cx  - d ||_2
    s.t.      Ax  = b
              Gx <= h

    This function is mainly meant for benchmarking the impact of using the L1
    instead of the L2 norm.
    """

    # Make sure the dimensionalities of the input matrices are correct
    C, d, A, b, G, h = _check_basis_pursuit_params(C, d, A, b, G, h)

    # Compute the matrices for the QP problem
    Pqp = C.T @ C
    qqp = -C.T @ d

    # Solve the QP problem
    return _solve_qp(Pqp, qqp, G, h, A, b, tol=tol, disp=disp)


def solve_weights_qp(A,
                     b,
                     valid=None,
                     iTh=0.0,
                     nonneg=True,
                     reg=DEFAULT_REG,
                     tol=DEFAULT_TOL):
    """
    Same as np.linalg.lstsq but uses cvxopt and adds regularisation.
    Additionally allows to solve for negative target values using an
    inequality condition instead of an "equals" condition by marking positive
    samples as "valid".
    """

    #
    # Step 1: Count stuff and setup indices used to partition the matrices
    #         into smaller parts.
    #

    # Compute the number of slack variables required to solve this problem
    n_cstr, n_vars = A.shape
    n_cstr_valid = int(np.sum(valid))
    n_cstr_invalid = n_cstr - int(np.sum(valid))
    n_slack = n_cstr_invalid
    n_vars_total = n_vars + n_slack

    # Variables
    v0 = 0
    v1 = n_vars
    v2 = v1 + n_slack

    # Quadratic constraints
    a0 = 0
    a1 = a0 + n_cstr_valid
    a2 = a1 + n_vars
    a3 = a2 + n_slack

    # Inequality constraints
    g0 = 0
    g1 = g0 + n_cstr_invalid
    g2 = g1 + (n_vars if nonneg else 0)

    #
    # Step 2: Assemble the QP matrices
    #

    # We need balance the re-weight error for the super- (valid) and
    # sub-threshold (invalid) constraints. This is done by dividing by the
    # number of valid/invalid constraints. We need to multiply with the number
    # of constraints since the regularisation factor has been chosen in such a
    # way that the errors are implicitly divided by the number of constraints.
    m1 = np.sqrt(n_cstr / max(1, n_cstr_valid))
    m2 = np.sqrt(n_cstr / max(1, n_cstr_invalid))

    # Copy the valid constraints to Aext
    Aext = np.zeros((a3, n_vars_total))
    bext = np.zeros(a3)
    Aext[a0:a1, v0:v1] = A[valid] * m1
    bext[a0:a1] = b[valid] * m1

    # Regularise the weights
    Aext[a1:a2, v0:v1] = np.eye(n_vars) * np.sqrt(reg)

    # Penalise slack variables
    Aext[a2:a3, v1:v2] = np.eye(n_slack) * m2

    # Form the inequality constraints for the matrices G and h
    G = np.zeros((g2, n_vars_total))
    G[g0:g1, v0:v1] = A[~valid]
    G[g0:g1, v1:v2] = -np.eye(n_slack)
    G[g1:g2, v0:v1] = -np.eye(n_vars) if nonneg else 0.0
    h = np.zeros(G.shape[0])
    h[g0:g1] = iTh

    #
    # Step 3: Solve the QP
    #

    x = solve_linearly_constrained_quadratic_loss(
        Aext, bext, None, None, G, h, tol=tol)
    return x[:n_vars, 0]

class SolverTask(collections.namedtuple('SolverTask', [
        'Apre', 'Jpost', 'w', 'connectivity', 'iTh', 'nonneg', 'renormalise',
        'tol', 'reg', 'use_lstsq', 'valid', 'i', 'n_samples'
    ])):
    pass

def _solve_single(t):
    # Fetch the excitatory and inhibitory pre-neurons
    exc, inh = t.connectivity
    Npre_exc = np.sum(exc)
    Npre_inh = np.sum(inh)
    Npre_tot = Npre_exc + Npre_inh

    # Just abort if there ist nothing to do
    if Npre_tot == 0:
        return t.i, np.zeros((2, 0))

    # Renormalise the target currents to a maximum magnitude of one and adapt
    # the model weights accordingly. Since it holds
    #
    #             w[0] + w[1] * gE + w[2] * gI
    # J(gE, gI) = ---------------------------- ≈ Jpost
    #             w[3] + w[4] * gE + w[5] * gI
    #
    # scaling the first three or last three weight vector components will scale
    # the predicted target current. Scaling w[1], w[2], w[4], w[5] will re-scale
    # the magnitude of the synaptic weights

    # Determine the current scaling factor. This should be about 1e9 / A.
    if t.renormalise:
        # Determine all scaling factors
        Wscale = 1.0e-9
        Λscale = 1.0 / (t.w[1]**2
                        )  # Need to scale the regularisation factor as well

        # Compute synaptic weights in nS
        t.w[[1, 2, 4, 5]] *= Wscale

        # Set w[1]=1 for better numerical stability/conditioning
        t.w[...] /= t.w[1]
    else:
        Wscale, Λscale = 1.0, 1.0

    # Account for the number of samples
    Λscale *= t.n_samples

    # Demangle the model weight vector
    a0, a1, a2, b0, b1, b2 = t.w

    # Clip Jtar to the valid range.
    warning_msgs = []
    if np.abs(b2) > 0 and np.abs(b1) > 0:
        if (a1 / b1) < np.max(t.Jpost):
            warning_msgs.append(
                ("Target currents for neuron {} cannot be reached! {:.3g} ∉ [" +
                 "{:.3g}, {:.3g}]")
            .format(t.i, np.max(t.Jpost), a2 / b2, a1 / b1))
        t.Jpost[...] = t.Jpost.clip(0.975 * a2 / b2, 0.975 * a1 / b1)

    # Split the pre activities into neurons marked as excitatory,
    # as well as neurons marked as inhibitory
    Apre_exc, Apre_inh = t.Apre[:, exc], t.Apre[:, inh]

    # Assemble the "A" and "b" matrices
    A = np.concatenate((
        np.diag(a1 - b1 * t.Jpost) @ Apre_exc,
        np.diag(a2 - b2 * t.Jpost) @ Apre_inh,
    ),  axis=1)
    b = t.Jpost * b0 - a0

    # Solve the least-squares problem, either using QP (including the
    # sub-threshold inequality/"mask_negative") or the lstsq/nnls functions.
    if not t.use_lstsq:
        fws = solve_weights_qp(
            A,
            b,
            valid=t.valid,
            nonneg=t.nonneg,
            iTh=t.iTh * b0 - a0,  # Transform iTh in the same way as the target currents
            reg=t.reg * Λscale,
            tol=t.tol)
    else:
        # Compute Γ and Υ
        Γ = A.T @ A + t.reg * Λscale * np.eye(A.shape[1])
        Υ = A.T @ b

        # Solve for weights using NNLS
        if t.nonneg:
            fws = scipy.optimize.nnls(Γ, Υ, maxiter=10*t.n_samples)[0]
        else:
            fws = np.linalg.lstsq(Γ, Υ, rcond=None)[0]

    return t.i, fws[:Npre_exc] * Wscale, fws[Npre_exc:] * Wscale, warning_msgs


def solve(Apre,
          Jpost,
          ws,
          connectivity=None,
          iTh=None,
          nonneg=True,
          renormalise=True,
          tol=None,
          reg=None,
          use_lstsq=False):
    # Set some default values
    if tol is None:
        tol = DEFAULT_TOL
    if reg is None:
        reg = DEFAULT_REG

    # Fetch some counts
    assert Apre.shape[0] == Jpost.shape[0]
    m = Apre.shape[0]
    Npre = Apre.shape[1]
    Npost = Jpost.shape[1]

    # Use an all-to-all connection if connectivity is set to None
    if connectivity is None:
        connectivity = np.ones((2, Npre, Npost), dtype=np.bool)

    # Create a neuron model parameter vector for each neuron, if the parameters
    # are not already in this format
    assert ws.size == 6 or ws.ndim == 2, "Model weight vector must either be 6-element one-dimensional or a 2D matrix"
    if (ws.size == 6):
        ws = np.repeat(ws.reshape(1, -1), Npost, axis=0)
    else:
        assert ws.shape[0] == Npost and ws.shape[1] == 6, "Invalid model weight matrix shape"

    # Mark all samples as "valid" if valid is None, otherwise select those where
    # the post-current is larger than the threshold current
    if iTh is None:
        valid = np.ones((m, Npost), dtype=np.bool)
        iTh = 0.0
    else:
        valid = Jpost >= iTh

    # Iterate over each post-neuron individually and solve for weights. Do so
    # in parallel.
    tasks = [SolverTask(
        Apre, Jpost[:, i], ws[i], connectivity[:, :, i], iTh,
        nonneg, renormalise, tol, reg, use_lstsq,
        valid[:, i], i, m) for i in range(Npost)]
    WE, WI = np.zeros((2, Npre, Npost))
    with EnvGuard({"OMP_NUM_THREADS": "1"}) as env:
        ctx = multiprocessing.get_context("spawn")
        with ctx.Pool() as pool:
            for i, we, wi, warning_msgs in pool.imap_unordered(_solve_single, tasks):
                for msg in warning_msgs:
                    print('Warning: {}'.format(msg))
                exc, inh = connectivity[:, :, i]
                WE[exc, i], WI[inh, i] = we, wi

    return WE, WI




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from nengo.exceptions import ValidationError
from nengo.neurons import NeuronType
from nengo.params import IntParam, NumberParam
from nengo.dists import Uniform
from nengo.cache import Fingerprint

from nengo_bio.common import Excitatory, Inhibitory
from nengo_bio.internal import (lif_utils, multi_compartment_lif_parameters)


class MultiChannelNeuronType(NeuronType):
    inputs = ()

    def step_math(self, dt, J, output):
        """
        The step_math function is not used by MultiInputNeuron instances. This
        is mainly to prevent mixing of MultiInputNeuron instances with standard
        Nengo ensembles.
        """
        raise RuntimeError(
            "Neurons of type can only be used in conjunction with"
            "nengo_bio.Connection")

    @property
    def n_inputs(self):
        return len(self.inputs)

    def tune(self, dt, model, ens):
        return None

    def compile(self, dt, n_neurons, tuning):
        """
        Generates an efficient simulator for this neuron type.

        Returns a callable "step_math" function. The "step_math" function
        receives the current input and returns the output (whether a neuron
        spiked or not), as well as the current state.

        Parameters
        ----------
        dt : float
            The timestep that is going to be used in the simulation. Since this
            parameter is known ahead of time it can be treated as a constant
            expression, potentially increasing performance.
        n_neurons : int
            The number of neurons for which the function should be compiled.
        """
        raise NotImplementedError(
            "MultiInputNeuronType must implement the \"compile\" function.")


class LIF(MultiChannelNeuronType):
    """
    A standard single-compartment LIF neuron with separate exciatory and
    inhibitory inputs.
    """

    inputs = (Excitatory, Inhibitory)

    C_som = NumberParam('C', low=0, low_open=True)
    g_leak_som = NumberParam('g_leak', low=0, low_open=True),

    tau_spike = NumberParam('tau_spike', low=0)
    tau_ref = NumberParam('tau_ref', low=0)

    v_spike = NumberParam('v_spike')
    v_reset = NumberParam('v_reset')
    v_th = NumberParam('v_reset')

    E_rev_leak = NumberParam('E_leak')

    subsample = IntParam('subsample', low=1)

    def __init__(self,
                 C_som=1e-9,
                 g_leak_som=50e-9,
                 E_rev_leak=-65e-3,
                 tau_ref=2e-3,
                 tau_spike=1e-3,
                 v_th=-50e-3,
                 v_reset=-65e-3,
                 v_spike=20e-3,
                 subsample=10):

        super(LIF, self).__init__()

        self.C_som = C_som
        self.g_leak_som = g_leak_som
        self.E_rev_leak = E_rev_leak
        self.tau_ref = tau_ref
        self.tau_spike = tau_spike
        self.v_th = v_th
        self.v_reset = v_reset
        self.v_spike = v_spike
        self.subsample = subsample

    @property
    def threshold_current(self):
        return (self.v_th - self.v_reset) * self.g_leak_som

    def _lif_parameters(self):
        """
        Returns the LIF parameters of the somatic compartments. These parameters
        are used in the gain/bias computations.
        """
        tau_ref = self.tau_spike + self.tau_ref
        tau_rc = self.C_som / self.g_leak_som
        i_th = self.threshold_current
        return tau_ref, tau_rc, i_th

    def _lif_rate(self, J):
        """
        Returns the LIF rate for a given input current.
        """
        tau_ref, tau_rc, i_th = self._lif_parameters()
        return lif_utils.lif_rate(J / i_th, tau_ref, tau_rc)

    def _lif_rate_inv(self, a):
        """
        Returns the input current resulting in the given rate.
        """
        tau_ref, tau_rc, i_th = self._lif_parameters()
        return lif_utils.lif_rate_inv(a, tau_ref, tau_rc) * i_th

    def gain_bias(self, max_rates, intercepts):
        # Make sure the input is a 1D array
        max_rates = np.array(max_rates, dtype=float, copy=False, ndmin=1)
        intercepts = np.array(intercepts, dtype=float, copy=False, ndmin=1)

        # Make sure the maximum rates are not surpassing the maximally
        # attainable rate
        tau_ref, _, i_th = self._lif_parameters()
        inv_tau_ref = 1. / tau_ref if tau_ref > 0. else np.inf
        if np.any(max_rates > inv_tau_ref):
            raise ValidationError(
                "Max rates must be below the inverse "
                "of the sum of the refractory and spike "
                "period ({:0.3f})".format(inv_tau_ref),
                attr='max_rates',
                obj=self)

        # Solve the following linear system for gain, bias
        #   i_th  = gain * intercepts + bias
        #   i_max = gain              + bias
        i_max = self._lif_rate_inv(max_rates)
        gain = (i_max - i_th) / (1. - intercepts)
        bias = i_max - gain

        return gain, bias

    def max_rates_intercepts(self, gain, bias):
        # The max rate is defined as the rate for the input current gain + bias
        max_rates = self._lif_rate(gain + bias)

        # Solve i_th = gain * intercept + bias for the intercept; warn about
        # invalid values
        i_th = self.threshold_current
        intercepts = (i_th - bias) / gain
        if not np.all(np.isfinite(intercepts)):
            warnings.warn("Non-finite values detected in `intercepts`; this "
                          "probably means that `gain` was too small.")

        return max_rates, intercepts

    def rates(self, x, gain, bias):
        return self._lif_rate(gain * x + bias)

    def _params_som(self):
        return multi_compartment_lif_parameters.SomaticParameters(
            tau_ref=self.tau_ref,
            tau_spike=self.tau_spike,
            v_th=self.v_th,
            v_reset=self.v_reset,
            v_spike=self.v_spike,
        )

    def _params_den(self):
        return multi_compartment_lif_parameters.DendriticParameters.\
            make_lif(
            C_som=self.C_som,
            g_leak_som=self.g_leak_som,
            E_rev_leak=self.E_rev_leak)

    def _compile(self, dt, force_python_sim=False):
        import nengo_bio.internal.multi_compartment_lif_cpp as mcl_cpp
        import nengo_bio.internal.multi_compartment_lif_python as mcl_python

        # Either instantiate the C++ simulator or the reference simulator
        params_som = self._params_som()
        params_den = self._params_den()
        if force_python_sim or not mcl_cpp.supports_cpp():
            sim_class = mcl_python.compile_simulator_python(
                params_som, params_den, dt=dt, ss=self.subsample)
        else:
            sim_class = mcl_cpp.compile_simulator_cpp(
                params_som, params_den, dt=dt, ss=self.subsample)

        # Return the simulator class
        return sim_class

    def _estimate_model_weights(self):
        return np.array((0., 1., -1., 1., 0, 0))

    def _filter_input(self, in_exc, in_inh):
        """
        Function used to quickly discard samples that will definetively lead to
        a zero output rate.
        """
        b0, b1, b2, a0, a1, a2 = self._estimate_model_weights()
        return (b0 + b1 * in_exc + b2 * in_inh) / (
            a0 + a1 * in_exc + a2 * in_inh) * 1.2 > self.threshold_current

    def _estimate_input_range(self, max_rate):
        """
        This function returns the 2D area over the excitatory and inhibitory
        input that should be sampled by the "tune" function.
        """

        # Fetch the model parameters. Estimate the absolute maximum and minimum
        # current.
        b0, b1, b2, a0, a1, a2 = self._estimate_model_weights()
        Jmin = b2 / a2 if a2 != 0.0 else -np.inf
        Jmax = b1 / a1 if a1 != 0.0 else  np.inf

        # Convert the given ramp to a current. Clamp the rate to the maximum/
        # minimum currents.
        J = self._lif_rate_inv(max_rate)
        J = np.clip(J, Jmin * 0.95, Jmax * 0.95)

        # Compute the gE that will reach the computed J for gI = 0
        gE = -(b0 - J * a0) / (b1 - J * a1)

        # Compute the gI that will result in J = i_th for the above gE
        gI = (self.threshold_current - (b0 + b1 * gE)) / b2

        # Return gE and gI
        return gE, gI

    def tune(self, dt, model, ens):
        from nengo_bio.internal.multi_compartment_lif_sim import GaussianSource
        from nengo_bio.internal.model_weights import tune_two_comp_model_weights

        # Fetch the maximum rates for which we need to determine the neuron
        # parameters
        max_rates = model.params[ens].max_rates
        if isinstance(ens.max_rates, Uniform):
            min_max_rate = ens.max_rates.low
            max_max_rate = ens.max_rates.high
        else:
            min_max_rate = np.min(max_rates)
            max_max_rate = np.max(max_rates)

        # Get the parameter hash
        params_hash = str(Fingerprint([self, dt]))

        # Function running a single neuron simulation
        sim_class = self._compile(dt)

        def run_single_sim(idx, out, in_exc, in_inh):
            xs = np.asarray((in_exc, in_inh), order='C', dtype=np.float64)
            # TODO: Get these parameters from the connection
            sim_class().run_single_with_gaussian_sources(
                out, [
                    GaussianSource(
                        seed=4902 + idx,
                        mu=in_exc,
                        sigma=in_exc * 0.2,
                        tau=5e-3,
                        offs=0.0),
                    GaussianSource(
                        seed=5821 + 7 * idx,
                        mu=in_inh,
                        sigma=in_inh * 0.2,
                        tau=5e-3,
                        offs=0.0),
                ])

        return tune_two_comp_model_weights(
            dt=dt,
            max_rates=max_rates,
            min_max_rate=min_max_rate,
            max_max_rate=max_max_rate,
            run_single_sim=run_single_sim,
            estimate_input_range=self._estimate_input_range,
            filter_input=self._filter_input,
            lif_rate_inv=self._lif_rate_inv,
            params_hash=params_hash)

    def compile(self,
                dt,
                n_neurons,
                tuning=None,
                force_python_sim=False,
                get_class=False):
        sim_class = self._compile(dt, force_python_sim)
        if get_class:
            return sim_class
        return sim_class(n_neurons).run_step_from_memory


class LIFCond(LIF):
    """
    Single compartment LIF neuron with conductance-based synapses.
    """

    inputs = (Excitatory, Inhibitory)

    E_rev_exc = NumberParam('E_exc')
    E_rev_inh = NumberParam('E_inh')

    def __init__(self,
                 C_som=1e-9,
                 g_leak_som=50e-9,
                 E_rev_leak=-65e-3,
                 E_rev_exc=20e-3,
                 E_rev_inh=-75e-3,
                 tau_ref=2e-3,
                 tau_spike=1e-3,
                 v_th=-50e-3,
                 v_reset=-65e-3,
                 v_spike=20e-3,
                 subsample=10):

        super(LIFCond, self).__init__(
            C_som=C_som,
            g_leak_som=g_leak_som,
            E_rev_leak=E_rev_leak,
            tau_ref=tau_ref,
            tau_spike=tau_spike,
            v_th=v_th,
            v_reset=v_reset,
            v_spike=v_spike,
            subsample=subsample)

        self.E_rev_exc = E_rev_exc
        self.E_rev_inh = E_rev_inh

    def _estimate_model_weights(self):
        v_som = 0.5 * (self.v_th + self.v_reset)
        ws = np.array((
            0.,
            (self.E_rev_exc - v_som),
            (self.E_rev_inh - v_som),
            1.,
            0.,
            0.,
        ))

        # Normalise ws[1] = 1
        ws = ws / ws[1]

        return ws

    def _params_den(self):
        return multi_compartment_lif_parameters.DendriticParameters.\
            make_lif_cond(
            C_som=self.C_som,
            g_leak_som=self.g_leak_som,
            E_rev_leak=self.E_rev_leak,
            E_rev_exc=self.E_rev_exc,
            E_rev_inh=self.E_rev_inh
        )


class TwoCompLIF(LIFCond):
    """
    A two-compartment LIF neuron with conductance based synapses.

    A TwoCompLIF neuron consists of a somatic as well as a dendritic
    compartment.
    """

    inputs = (Excitatory, Inhibitory)

    C_den = NumberParam('c_den', low=0, low_open=True)

    g_leak_den = NumberParam('g_leak_den', low=0, low_open=True)
    g_couple = NumberParam('g_couple', low=0, low_open=True)

    def __init__(self,
                 C_som=1e-9,
                 C_den=1e-9,
                 g_leak_som=50e-9,
                 g_leak_den=50e-9,
                 g_couple=50e-9,
                 E_rev_leak=-65e-3,
                 E_rev_exc=20e-3,
                 E_rev_inh=-75e-3,
                 tau_ref=2e-3,
                 tau_spike=1e-3,
                 v_th=-50e-3,
                 v_reset=-65e-3,
                 v_spike=20e-3,
                 subsample=10):

        super(TwoCompLIF, self).__init__(
            C_som=C_som,
            g_leak_som=g_leak_som,
            E_rev_leak=E_rev_leak,
            E_rev_exc=E_rev_exc,
            E_rev_inh=E_rev_inh,
            tau_ref=tau_ref,
            tau_spike=tau_spike,
            v_th=v_th,
            v_reset=v_reset,
            v_spike=v_spike,
            subsample=subsample)

        self.C_den = C_den
        self.g_leak_den = g_leak_den
        self.g_couple = g_couple

    def _estimate_model_weights(self):
        v_som = 0.5 * (self.v_th + self.v_reset)
        ws = np.array((
            self.g_couple * self.g_leak_den * (self.E_rev_leak - v_som),
            self.g_couple * (self.E_rev_exc - v_som),
            self.g_couple * (self.E_rev_inh - v_som),
            self.g_couple + self.g_leak_den,
            1.,
            1.,
        ))

        # Normalise ws[1] = 1
        ws = ws / ws[1]

        return ws

    def _params_den(self):
        return multi_compartment_lif_parameters.DendriticParameters.\
            make_two_comp_lif(
            C_som=self.C_som,
            C_den=self.C_den,
            g_leak_som=self.g_leak_som,
            g_leak_den=self.g_leak_den,
            g_couple=self.g_couple,
            E_rev_leak=self.E_rev_leak,
            E_rev_exc=self.E_rev_exc,
            E_rev_inh=self.E_rev_inh
        )


# Alias for TwoCompLIF emphasizing the use of conductance based synapses
TwoCompLIFCond = TwoCompLIF

# Whitelist the neuron types
Fingerprint.whitelist(LIF)
Fingerprint.whitelist(LIFCond)
Fingerprint.whitelist(TwoCompLIF)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from nengo_bio.common import steal_param, Excitatory, Inhibitory
from nengo_bio.solvers import SolverWrapper, QPSolver

from nengo.params import Parameter, BoolParam, IntParam, StringParam, \
                         Default, Unconfigurable

import nengo.base
import nengo.config
import nengo.connection
import nengo.dists
import nengo.exceptions
import nengo.synapses
import nengo.builder

class ConnectionPart(nengo.connection.Connection):

    dales_principle = True

    def __init__(self, *args, **kw_args):
        self._synapse_type = steal_param('synapse_type', kw_args, Default)
        super(ConnectionPart, self).__init__(*args, **kw_args)

    @property
    def synapse_type(self):
        return self._synapse_type

    @property
    def kind(self):
        return str(self._synapse_type)

class MultiEnsemble(nengo.base.SupportDefaultsMixin):
    """
    The MultiEnsemble class represents a view at a list of/set of ensembles.
    Given a descriptor `descr` the represented value of the ensembles is either
    stacked, or joined. "Stack" operations are denoted as a tuple

        (ens_a, ens_b)

    where ens_a is a 1D ensemble and ens_b is a 2D ensemble will result in a 3D
    multi ensemble. "Join" operations are denoted as a set:

        {ens_a, ens_b}

    Here, both ens_a and ens_b have to have the same dimensionality -- the
    resulting multi-ensemble will provide a view on ens_a and ens_b where ens_a
    and ens_b represent the same value at the same time.
    """

    OP_NONE = 0
    OP_STACK = 1
    OP_JOIN = 2

    n_neurons = IntParam('n_neurons', low=1, readonly=True)
    dimensions = IntParam('dimensions', low=1, readonly=True)

    def __init__(self, descr, operator=None):
        # Nengo objects that can be base cases for the multi ensembles
        SUPPORTED_NENGO_OBJS = (
            nengo.ensemble.Ensemble
        )

        # Determine the operator type depending on whether the given descriptor
        # is a tuple, set, or just an ensemble
        if isinstance(descr, SUPPORTED_NENGO_OBJS) and \
           ((operator is None) or (operator == OP_NONE)):
            self.operator = MultiEnsemble.OP_NONE
            self.objs = (descr,)
        elif isinstance(descr, tuple) and (operator is None):
            self.operator = MultiEnsemble.OP_STACK
            self.objs = descr
        elif isinstance(descr, set) and (operator is None):
            self.operator = MultiEnsemble.OP_JOIN
            self.objs = descr
        elif isinstance(descr, MultiEnsemble):
            if not (operator is None):
                raise ValueError(
                    "\"operator\" must be None when initialising a "
                    "MultiEnsemble with a MultiEnsemble.")
            self.operator = descr.operator
            self.objs = descr.objs
        elif (operator in [MultiEnsemble.OP_STACK,
                           MultiEnsemble.OP_JOIN]) and hasattr(descr, '__len__'):
            self.operator = operator
            self.objs = tuple(descr)
        else:
            raise ValueError(
                "Pre-object must either be a tuple, set, or an Ensemble")

        # Recursively turn the individual objects into MultiEnsemble objects
        if self.operator != MultiEnsemble.OP_NONE:
            self.objs = tuple(map(MultiEnsemble, self.objs))

        # In case this is a "join" operator, make sure the objects have the same
        # dimensionality
        if self.operator == MultiEnsemble.OP_JOIN:
            if len(set(map(lambda x: x.size_out, self.objs))) > 1:
                raise ValueError(
                    "Ensembles must have the same dimensionality to be joined.")

        # Accumulate the number of dimensions and neurons
        self.dimensions = self._get_accu_dim_attr('dimensions')
        self.n_neurons = sum(map(lambda x: x.n_neurons, self.objs))

    def _get_accu_dim_attr(self, attr):
        """
        Internal function to accumulate a quantity with the name `attr`
        according to whether this ensemble implements a stack or a join
        operation.
        """
        if self.operator == MultiEnsemble.OP_JOIN:
            return 0 if len(self.objs) == 0 else getattr(self.objs[0], attr)
        else:
            return sum(map(lambda x: getattr(x, attr), self.objs))

    def __len__(self):
        return self.dimensions

    def __repr__(self):
        if self.operator == MultiEnsemble.OP_NONE:
            return str(self.objs[0])
        else:
            d0 = '{' if self.operator == MultiEnsemble.OP_JOIN else '('
            d1 = '}' if self.operator == MultiEnsemble.OP_JOIN else ')'
            return "{}{}{}".format(d0, ', '.join(map(str, self.objs)), d1)

    @property
    def size_in(self):
        return self._get_accu_dim_attr("size_in")

    @property
    def size_out(self):
        return self._get_accu_dim_attr("size_out")

    def flatten(self):
        """
        Returns a flat tuple of ensembles, a tuple containing the 1D activity
        slice for each ensemble (assigning a neuron ID to each neuron), and a
        tuple containing the 2D eval_points slice.
        """

        if self.operator == MultiEnsemble.OP_NONE:
            ens = self.objs
            ns = (slice(0, self.n_neurons),) # neuron indices
            ds = (slice(0, self.dimensions),) # dimension indicies
        elif (self.operator == MultiEnsemble.OP_STACK) or\
             (self.operator == MultiEnsemble.OP_JOIN):
            arr_ens, arr_ns, arr_ds = [], [], []
            nn, dn, en = 0, 0, 0
            for i, obj in enumerate(self.objs):
                ens, ns, ds = obj.flatten()

                # Increment the neuron numbers by nn
                ns = tuple(slice(nn + x.start, nn + x.stop) for x in ns)
                nn = ns[-1].stop

                # Increment the dimension/eval points index
                ds = tuple(slice(dn + x.start, dn + x.stop) for x in ds)
                if self.operator == MultiEnsemble.OP_STACK:
                    dn = ds[-1].stop

                # Append the lists to the arrays
                arr_ens.append(ens); arr_ns.append(ns); arr_ds.append(ds)

            # Merge the arrays into a single tuple
            ens, ns, ds = sum(arr_ens, ()), sum(arr_ns, ()), sum(arr_ds, ())
        return ens, ns, ds


class PreParam(Parameter):
    """
    The PreParam class is used by nengo_bio.Connection to describe the list
    of pre-objects that are involved in a certain connection.
    """

    def __init__(self, name):
        super().__init__(name, default=Unconfigurable,
                         optional=False, readonly=True)

    def coerce(self, instance, nengo_obj):
        # Try to convert the given ensemble into a MultiEnsemble
        try:
            obj = MultiEnsemble(nengo_obj)
        except ValueError as e:
            raise nengo.exceptions.ValidationError(
                e.msg, attr=self.name, obj=instance)

        return super().coerce(instance, obj)


class ConnectionFunctionParam(nengo.connection.ConnectionFunctionParam):
    """Connection-specific validation for functions."""

    coerce_defaults = False

    def check_function_can_be_applied(self, conn, function_info):
        function, size = function_info
        type_pre = type(conn.pre_obj).__name__

# Nengo 2.8 compatibility
if hasattr(nengo.connection, 'ConnectionTransformParam'):
    ConnectionTransformParam = nengo.connection.ConnectionTransformParam
else:
    ConnectionTransformParam = nengo.connection.TransformParam


class Connection(nengo.config.SupportDefaultsMixin):

    label = StringParam(
        'label', default=None, optional=True)
    seed = IntParam(
        'seed', default=None, optional=True)

    pre = PreParam(
        'pre')
    post = nengo.connection.PrePostParam(
        'post', nonzero_size_in=True)

    synapse_exc = nengo.connection.SynapseParam(
        'synapse_exc', default=nengo.synapses.Lowpass(tau=0.005))
    synapse_inh = nengo.connection.SynapseParam(
        'synapse_inh', default=nengo.synapses.Lowpass(tau=0.005))

    function_info = ConnectionFunctionParam(
        'function', default=None, optional=True)

    transform = ConnectionTransformParam(
        'transform', default=1.0)

    solver = nengo.solvers.SolverParam(
        'solver', default=QPSolver())
    eval_points = nengo.dists.DistOrArrayParam(
        'eval_points', default=None, optional=True, 
        sample_shape=('*', 'size_in'))
    scale_eval_points = BoolParam(
        'scale_eval_points', default=True)
    n_eval_points = IntParam(
        'n_eval_points', default=None, optional=True)
    decode_bias = BoolParam(
        'decode_bias', default=True)

    max_n_post_synapses = IntParam(
        'max_n_post_synapses', low=0, default=None, optional=True)
    max_n_post_synapses_exc = IntParam(
        'max_n_post_synapses_exc', low=0, default=None, optional=True)
    max_n_post_synapses_inh = IntParam(
        'max_n_post_synapses_inh', low=0, default=None, optional=True)

    _param_init_order = [
        'pre', 'post', 'synapse_exc', 'synapse_inh', 'function_info'
    ]

    def __init__(self, pre, post,
                 synapse_exc=Default,
                 synapse_inh=Default,
                 function=Default,
                 transform=Default,
                 solver=Default,
                 eval_points=Default,
                 scale_eval_points=Default,
                 n_eval_points=Default,
                 decode_bias=Default,
                 max_n_post_synapses=Default,
                 max_n_post_synapses_exc=Default,
                 max_n_post_synapses_inh=Default,
                 label=Default,
                 seed=Default):
        super().__init__()

        # Copy the parameters
        self.label = label
        self.seed = seed

        self.pre = pre
        self.post = post
        self.synapse_exc = synapse_exc
        self.synapse_inh = synapse_inh
        self.eval_points = eval_points
        self.scale_eval_points = scale_eval_points
        self.n_eval_points = n_eval_points
        self.decode_bias = decode_bias
        self.function_info = function
        self.transform = transform
        self.solver = solver

        self.max_n_post_synapses = max_n_post_synapses
        self.max_n_post_synapses_exc = max_n_post_synapses_exc
        self.max_n_post_synapses_inh = max_n_post_synapses_inh

        # For each pre object add two actual nengo connections: an excitatory
        # path and an inhibitory path
        self.connections = []
        arr_ens, arr_ns, _ = self.pre.flatten()
        for i, (ens, ns) in enumerate(zip(arr_ens, arr_ns)):
            def mkcon(synapse_type, synapse):
                return ConnectionPart(
                    pre=ens,
                    post=self.post,
                    transform=np.zeros((self.post.size_in, ens.size_out)),
                    seed=self.seed,
                    synapse=synapse,
                    solver=SolverWrapper(
                        self.solver, i, self, ns, synapse_type),
                    synapse_type=synapse_type)
            self.connections.append((
                mkcon(Excitatory, synapse_exc),
                mkcon(Inhibitory, synapse_inh)))

    def __str__(self):
        return "<Connection {}>".format(self._str)

    def __repr__(self):
        return "<Connection at {:06X} {}>".format(id(self), self._str)

    @property
    def _str(self):
        if self.label is not None:
            return self.label

        desc = "" if self.function is None else " computing '{}'".format(
            function_name(self.function))
        return "from {} to {}{}".format(self.pre, self.post, desc)

    @property
    def function(self):
        return self.function_info.function

    @function.setter
    def function(self, function):
        self.function_info = function

    @property
    def is_decoded(self):
        return not (self.solver.weights or (
            isinstance(self.pre_obj, Neurons)
            and isinstance(self.post_obj, Neurons)))

    @property
    def _label(self):
        if self.label is not None:
            return self.label

        return "from %s to %s%s" % (
            self.pre, self.post,
            " computing '%s'" % function_name(self.function)
            if self.function is not None else "")

    @property
    def post_obj(self):
        return self.post

    @property
    def pre_obj(self):
        return self.pre

    @property
    def pre_slice(self):
        return slice(None)

    @property
    def post_slice(self):
        return slice(None)

    @property
    def size_in(self):
        return self.pre.size_out

    @property
    def size_mid(self):
        size = self.function_info.size
        return self.size_in if size is None else size

    @property
    def size_out(self):
        return self.post.size_in




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import warnings
import numpy as np

from nengo_bio.common import Excitatory, Inhibitory
from nengo_bio.connection import ConnectionPart, Connection
from nengo_bio.solvers import SolverWrapper, ExtendedSolver
from nengo_bio.neurons import MultiChannelNeuronType

from nengo.exceptions import NengoWarning, BuildError
from nengo.ensemble import Ensemble, Neurons

import nengo.builder
from nengo.builder.operator import Copy, Reset

class BuiltConnection:
    def __init__(self):
        self.weights = {
            Excitatory: None,
            Inhibitory: None
        }

def get_multi_ensemble_eval_points(
    model, mens, n_eval_points=None, scale_eval_points=True, rng=np.random):
    """
    This function generates the evaluation points for the given MultiEnsemble.
    """

    def choice(A, n):
        return A[rng.randint(0, A.shape[0], n)]

    if mens.operator == mens.OP_NONE:
        # Recursion base case. The MultiEnsemble encapsulates a single Ensemble
        # instance -- just return the evaluation points associated with this
        # ensemble.
        pnts = model.params[mens.objs[0]].eval_points
        pnts.setflags(write=False)

        # Scale the evaluation points in case this was requested by the
        # connection
        if scale_eval_points:
            pnts = pnts * mens.objs[0].radius

        # In case a specific number of eval points is requested, create a random
        # selection of eval points
        if not n_eval_points is None:
            pnts = choice(pnts, n_eval_points)
        return pnts
    elif (mens.operator == mens.OP_STACK) or (mens.operator == mens.OP_JOIN):
        # For each MultiEnsemble object in the stack/join, fetch the evaluation
        # points associated with that MultiEnsemble. Track the maximum number
        # of evaluation points.
        pnts_per_obj, n_pnts = [None] * len(mens.objs), [0] * len(mens.objs)
        for i, obj in enumerate(mens.objs):
            pnts = get_multi_ensemble_eval_points(
                model, mens.objs[i], n_eval_points, scale_eval_points, rng)
            pnts_per_obj[i] = pnts
            n_pnts[i] = pnts.shape[0]
        max_n_pnts = max(n_pnts)

        # Either select n_eval_points or the maximum number of eval points
        # as the number of evaluation points to generate.
        if n_eval_points is None:
            n_eval_points = max_n_pnts

        if mens.operator == mens.OP_STACK:
            # Write the evaluation points to a contiguous array.
            pnts = np.empty((max_n_pnts, mens.dimensions))
            d = 0
            for i, p in enumerate(pnts_per_obj):
                n_pnts, n_dims = p.shape
                if n_eval_points >= max_n_pnts:
                    # Write the points to the resulting array, fill the
                    # remaining space with randomly selected samples
                    pnts[:n_pnts, d:(d+n_dims)] = p
                    pnts[n_pnts:, d:(d+n_dims)] = \
                        choice(p, n_eval_points - n_pnts)
                else:
                    # Ranomly select n_eval_points points and write them to
                    # the target array
                    pnts[:, d:(d+n_dims)] = choice(p, n_eval_points)

                # Increment the dimension counter
                d += n_dims
            return pnts
        elif mens.operator == mens.OP_JOIN:
            # Write the evaluation points to a contiguous array and select
            # max_n_pnts of those
            return choice(np.concatenate(pnts_per_obj, axis=0), n_eval_points)
    else:
        raise ValueError("Invalid MultiEnsemble operator")


def get_eval_points(model, conn, rng=np.random):
    # In case no eval_points object has been specified
    if conn.eval_points is None:
        return get_multi_ensemble_eval_points(
            model, conn.pre_obj, conn.n_eval_points, conn.scale_eval_points,
            rng)
    else:
        if conn.scale_eval_points:
            warnings.warn(NengoWarning(
                "Setting scale_eval_points to True has no effect on {} if "
                "eval_points are specified manually.".format(conn.pre_obj)))
        return nengo.builder.ensemble.gen_eval_points(
            conn, conn.eval_points, rng, False)


def get_multi_ensemble_activities(model, mens, eval_points):
    # Create the empty activity matrix
    n_eval_points = eval_points.shape[0]
    n_neurons = mens.n_neurons
    activities = np.empty((n_eval_points, n_neurons))

    # Iterate over all ensembles and ask Nengo to compute the activities. Write
    # the results to the activity matrix.
    arr_ens, arr_ns, arr_ds = mens.flatten()
    for ens, ns, ds in zip(arr_ens, arr_ns, arr_ds):
        activities[:, ns] = nengo.builder.ensemble.get_activities(
            model.params[ens], ens, eval_points[:, ds])

    return activities


def get_multi_ensemble_synapse_types(model, mens):
    # Create the empty synapse type map
    n_neurons = mens.n_neurons
    synapse_types = np.empty((2, n_neurons), dtype=np.bool)

    # Iterate over all ensembles and write the synapse types to the map
    arr_ens, arr_ns, _ = mens.flatten()
    for ens, ns in zip(arr_ens, arr_ns):
        for i, type_ in enumerate((Excitatory, Inhibitory)):
            built_ens = model.params[ens]
            if hasattr(built_ens, "synapse_types"):
                synapse_types[i, ns] = built_ens.synapse_types[type_]
            else:
                synapse_types[i, ns] = np.ones(ens.n_neurons, dtype=np.bool)
    return synapse_types


def get_connectivity(conn, synapse_types, rng=np.random):
    # Create some convenient aliases
    Npre, Npost = conn.pre_obj.n_neurons, conn.post_obj.n_neurons
    mps = conn.max_n_post_synapses
    mps_E, mps_I = conn.max_n_post_synapses_exc, conn.max_n_post_synapses_inh
    has_mps = not (mps is None)
    has_mps_E, has_mps_I = not (mps_E is None), not (mps_I is None)

    # If no restrictions were given, just allow all-to-all connections
    if not (has_mps or has_mps_E or has_mps_I):
        return np.array((
            np.tile(synapse_types[0, :, None], Npost),
            np.tile(synapse_types[1, :, None], Npost)),
            dtype=np.bool)
    if has_mps and has_mps_E and has_mps_I:
        raise BuildError(
            "Specifying max_n_post_synapses as well as both "
            "max_n_post_synapses_exc and max_n_post_synapses_inh is invalid.")

    # Limit mps_E and mps_I to mps
    if has_mps:
        if has_mps_E:
            mps_E = min(mps, mps_E)
        if has_mps_I:
            mps_I = min(mps, mps_I)

    connectivity = np.zeros((2, Npre, Npost), dtype=np.bool)
    for i_post in range(Npost):
        # Get the indices of possible excitatory connection sites
        i_exc, i_inh = np.where(synapse_types[0])[0], np.where(synapse_types[1])[0]
        n_exc, n_inh = i_exc.size, i_inh.size

        # Select mps_E excitatory/mps_I inhibitory connections
        i_exc_sel, i_inh_sel = np.zeros((2, 0), dtype=np.int32)
        if has_mps_E:
            i_exc_sel = rng.choice(i_exc, size=min(mps_E, n_exc), replace=False)
        if has_mps_I:
            i_inh_sel = rng.choice(i_inh, size=min(mps_I, n_inh), replace=False)

        # We're done if both has_mps_E and has_mps_I are true. Otherwise, select
        # more neurons up to mps.
        if not (has_mps_E and has_mps_I):
            # If no maximum number of synapses is set, set it to the maximum
            # number of synapses that are still available
            if not has_mps:
                mps_rem = n_exc + n_inh
            else:
                mps_rem = mps
            mps_rem = max(0, mps_rem - i_exc_sel.size - i_inh_sel.size)

            if has_mps_E: # Need to select inhibitory neurons
                i_inh_sel = rng.choice(
                    i_inh, size=min(mps_rem, n_inh), replace=False)
            elif has_mps_I: # Need to select excitatory neurons
                i_exc_sel = rng.choice(
                    i_exc, size=min(mps_rem, n_exc), replace=False)
            else: # Need to select both excitatory and inhibitory neurons
                idcs = rng.choice(
                    np.arange(0, n_inh + n_exc, dtype=np.int32),
                    size=mps_rem, replace=False)
                i_exc_sel = i_exc[idcs[idcs < n_exc]]
                i_inh_sel = i_inh[idcs[idcs >= n_exc] - n_exc]

        # Set the corresponding entries in the connectivity matrix to true
        connectivity[0, i_exc_sel, i_post] = True
        connectivity[1, i_inh_sel, i_post] = True
    return connectivity

def remove_bias_current(model, ens):
    if not 'bias' in model.sig[ens.neurons]:
        return

    sig_post_bias = model.sig[ens.neurons]['bias']
    sig_post_in = model.sig[ens.neurons]['in']
    for i, op in enumerate(model.operators):
        if isinstance(op, Copy):
            if (op.src is sig_post_bias) and (op.dst is sig_post_in):
                # Delete the copy operator and instead add a reset operator
                del model.operators[i]
                model.add_op((Reset(sig_post_in)))


@nengo.builder.Builder.register(SolverWrapper)
def build_solver(model, solver, _, rng, *args, **kwargs):
    # Fetch the high-level connection
    conn = solver.connection # Note: this is the nengo_bio.Connection object
                             # and NOT the nengo.Connection object

    # If the high-level connection object has not been built, build it
    if not conn in model.params:
        ### TODO: Move to build_connection
        model.params[conn] = built_connection = BuiltConnection()

        # Remove the bias current from the target ensemble
        if conn.decode_bias:
            remove_bias_current(model, conn.post_obj)
        elif isinstance(conn.post_obj.neuron_type, MultiChannelNeuronType):
            raise BuildError(
                "decode_bias=False on connection {} invalid for post objects "
                "with multi-channel neurons.".format(conn))

        # Fetch the evaluation points, activites, and synapse types for the
        # entire MultiEnsemble
        eval_points = get_eval_points(
            model, conn, rng)
        activities = get_multi_ensemble_activities(
            model, conn.pre_obj, eval_points)
        synapse_types = get_multi_ensemble_synapse_types(
            model, conn.pre_obj)

        # Fetch the target values in representation space
        targets = nengo.builder.connection.get_targets(conn, eval_points)

        # Transform the target values
        if hasattr(nengo.connection, 'Dense'): # Nengo 2.8 compat
            transform = conn.transform.sample(rng=rng)
        else:
            transform = conn.transform
        targets = np.dot(targets, transform.T)

        # For the target population, fetch the gains and biases
        built_post_ens = model.params[conn.post_obj]
        encoders = built_post_ens.encoders / conn.post.radius
        gain = built_post_ens.gain
        bias = built_post_ens.bias

        # Compute the target currents
        target_currents = (targets @ encoders.T) * gain
        if conn.decode_bias:
            target_currents += bias

        # Construct the connectivity matrix for this connection
        connectivity = get_connectivity(conn, synapse_types, rng)

        # LIF neuron model parameters
        tuning = None
        if hasattr(built_post_ens, 'tuning'):
            tuning = built_post_ens.tuning

        i_th = 1.0
        if hasattr(conn.post_obj.neuron_type, 'threshold_current'):
            i_th = conn.post_obj.neuron_type.threshold_current
        WE, WI = solver(activities, target_currents, connectivity, i_th, tuning, rng)

        # If we're not targeting a MultiChannelNeuronType there really isn't
        # a distinction between excitatory and inhibitory input weights. Hence,
        # we must negate the inhibitory weights
        if not isinstance(conn.post_obj.neuron_type, MultiChannelNeuronType):
            WI = -WI

        built_connection.weights[Excitatory] = WE
        built_connection.weights[Inhibitory] = WI
    else:
        built_connection = model.params[conn]

    W = np.copy(
        built_connection.weights[solver.synapse_type][solver.neuron_indices].T)

    return None, W, None

@nengo.builder.Builder.register(ConnectionPart)
def build_connection(model, conn):
    # Run the original build_connection. This will trigger the above
    # build_solver function
    nengo.builder.connection.build_connection(model, conn)

    # Fetch the ensemble the connection part is connected to. Abort if the
    # target is not an ensemble
    if isinstance(conn.post_obj, Neurons):
        ens = conn.post_obj.ensemble
    elif isinstance(conn.post_obj, Ensemble):
        ens = conn.post_obj
    else:
        return

    # Nothing to do if the post neuron type is not a MultiChannelNeuronType
    if not isinstance(ens.neuron_type, MultiChannelNeuronType):
        return

    # Make sure the connection is actually connection to the "neurons" instance
    # and not the ensembles directly
    assert model.sig[conn]['out'] == model.sig[conn.post_obj.neurons]['in']

    # Determine which channel to connect to
    channel_idx = ens.neuron_type.inputs.index(conn.synapse_type)

    # Fetch the corresponding signal
    dst = model.sig[conn.post_obj.neurons]['in_{}'.format(channel_idx)]

    # Search for the copy operator that connects the connection to the unused
    # "in" signal and delete it. Create a new operator that connects the
    # connection to the right neuron channel.
    for i, op in enumerate(model.operators):
        if isinstance(op, Copy):
            if (op.dst is model.sig[conn.post_obj.neurons]['in']):
                del model.operators[i]
                model.add_op(Copy(
                    op.src, dst, None, None, inc=True,
                    tag="{}.{}".format(conn, conn.kind)))
                break




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

def lif_rate(J, tau_ref=2e-3, tau_rc=20e-3):
    """
    Calculates the firing rate of a LIF neuron with current based synapses.
    Input to the model is the current J. All parameters may be numpy arrays.

    J: input current
    tau_ref: refractory period in seconds.
    tau_rc: membrane time constant in seconds.
    """

    mask = 1 * (J > 1)
    t_spike = -np.log1p(-mask * 1.0 / (mask * J + (1 - mask))) * tau_rc
    return mask / (tau_ref + t_spike)


def lif_rate_inv(r, tau_ref=2e-3, tau_rc=20e-3):
    """
    Calculates the firing rate of a LIF neuron with current based synapses.
    Input to the model is the current J. All parameters may be numpy arrays.

    r: input rate
    tau_ref: refractory period in seconds.
    tau_rc: membrane time constant in seconds.
    """
    mask = 1.0 * (r > 1e-6)
    return -mask / (np.exp((r * tau_ref - 1) / ((1.0 - mask) + r * tau_rc)) - 1)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from nengo.params import BoolParam, NumberParam
from nengo.solvers import Solver
from nengo_bio.internal import qp_solver

class ExtendedSolver(Solver):
    """
    Extended solver class used by nengo_bio. Passes neuron types and target
    currents to the solver routine. Always solves for a full weight matrix.
    Non-compositional.
    """

    compositional = False

    def __init__(self):
        super().__init__(weights=True)

    def __call__(self, A, J, synapse_types, rng=np.random):
        raise NotImplementedError("Solvers must implement '__call__'")


class SolverWrapper(ExtendedSolver):
    def __init__(self, solver, pre_idx, connection, neuron_indices,
                 synapse_type):
        super().__init__()
        self.solver = solver
        self.pre_idx = pre_idx
        self.connection = connection
        self.neuron_indices = neuron_indices
        self.synapse_type = synapse_type

    def __call__(self, A, J, connectivity, i_th, tuning, rng=np.random):
        return self.solver(A, J, connectivity, i_th, tuning, rng)


class QPSolver(ExtendedSolver):

    reg = NumberParam('reg', low=0)
    relax = BoolParam('relax')

    def __init__(self, reg=1e-3, relax=False):
        super().__init__()
        self.reg = reg
        self.relax = relax

    def __call__(self, A, J, connectivity, i_th, tuning, rng=np.random):
        # Neuron model parameters. For now we only support current-based LIF
        if tuning is None:
            ws = np.array((0.0, 1.0, -1.0, 1.0, 0.0, 0.0))
        else:
            ws = tuning

        # Determine the final regularisatio nparameter
        reg = (self.reg * np.max(A))**2 * A.shape[1]

        # If subthreshold relaxation is switched off, set the spike threshold
        # to "None"
        i_th = None if not self.relax else i_th

        # Use the faster NNLS solver instead of CVXOPT if we do not need
        # current relaxation
        use_lstsq = i_th is None

        return qp_solver.solve(A, J, ws, connectivity,
                               iTh=i_th, reg=reg, use_lstsq=use_lstsq)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import numpy as np


class SomaticParameters:
    """
    The SomaticParameters class describes the parameters of the active LIF
    compartment such as the refractory period, spike time and threshold
    potential.
    """

    def __init__(self,
                 tau_ref=2e-3,
                 tau_spike=1e-3,
                 v_th=-50e-3,
                 v_reset=-65e-3,
                 v_spike=20e-3):
        self.tau_ref = tau_ref
        self.tau_spike = tau_spike
        self.v_th = v_th
        self.v_reset = v_reset
        self.v_spike = v_spike


class DendriticParameters:
    """
    The DendriticParameters class describes the connectivity of the dendritic
    tree. In particular, the dynamics of the tree are given as

    dv/dt = [C + diag(a_const + A @ x)] @ v + [b_const + B @ x]
    """

    def __init__(self, A, a_const, B, b_const, C):
        # Copy the matrices/vectors
        self.A = np.asarray(A, order='C', dtype=np.float64)
        self.a_const = np.asarray(a_const, order='C', dtype=np.float64)
        self.B = np.asarray(B, order='C', dtype=np.float64)
        self.b_const = np.asarray(b_const, order='C', dtype=np.float64)
        self.C = np.asarray(C, order='C', dtype=np.float64)

        # Make sure they have the right sizes
        self.a_const = self.a_const.flatten()
        self.b_const = self.b_const.flatten()
        assert self.A.ndim == self.B.ndim == self.C.ndim == 2
        assert self.C.shape[0] == self.C.shape[1]
        assert self.a_const.size == self.A.shape[0] == self.C.shape[0]
        assert self.b_const.size == self.B.shape[0] == self.C.shape[0]
        assert self.A.shape[1] == self.B.shape[1]

    @property
    def n_comp(self):
        """Returns the number of compartments in the neuron model."""
        return self.C.shape[0]

    @property
    def n_inputs(self):
        """Returns the number of inputs feeding into the neuron model."""
        return self.A.shape[1]

    def _reduce_system(self, A, b, i, v):
        """
        Reduces given the system by clamping a certain compartment i to the
        given potential v. Returns three matrices corresponding to the reduced
        system A_red, b_red and the vector required to compute the current
        flowing into the clamped compartment.
        """

        # Abort if this is a one-compartment neuron
        n = A.shape[0]
        if n <= 1:
            return 0, 0, 0

        # Fetch the A sub-matrix, add the current caused by the clamped somatic
        # compartment
        sel = list(range(0, i)) + list(range(i + 1, n))
        A_red = A[np.ix_(sel, sel)]
        b_red = b[sel] + v * A[sel, i]
        return A_red, b_red

    def vEq_extreme(self, params_som=None):
        """Returns the absolute minimum and maximum potentials reachable in each
           compartment by setting a conductance input to infinity. Note that
           current inputs will"""

        # Initialize v_min and v_max to the resting potential
        A, b = self.C + np.diag(self.a_const), self.b_const
        v_eq = -np.linalg.solve(A, b)
        v_min, v_max = v_eq, v_eq

        # Iterate over all compartments and all inputs and see what happens when
        # we set one of the inputs to infinity.
        for i in range(self.n_comp):
            for j in range(self.n_inputs):
                for sign in [1, -1]:
                    # Skip inputs that are not directly connected to the given
                    # compartment
                    if self.A[i, j] == 0.0 and self.B[i, j] == 0.0:
                        continue

                    # Compute the reversal potential for each compartment this input
                    # is connected to. Set current based inputs to a reversal
                    # potential of infinity
                    if self.A[i, j] == 0:
                        E_rev = sign * np.inf
                    else:
                        E_rev = self.B[i, j] / self.A[i, j]

                    # Assume the compartment had the potential we just computed. Do
                    # this by reducing system to a smaller one without this
                    # compartment.
                    if self.n_comp > 1:
                        A_red, b_red = self._reduce_system(A, b, i, E_rev)
                        v_eq = -np.linalg.solve(A_red, b_red)
                    else:
                        v_eq = E_rev

                    # Update the minimum/maximum potential
                    v_min = np.minimum(v_min, v_eq)
                    v_max = np.maximum(v_max, v_eq)

        # If present, use some knowledge from params_som to restrict the
        # minimum/maximum voltages further
        if not params_som is None:
            if np.isinf(v_min[0]):
                v_min[0] = params_som.v_reset
            if np.isinf(v_max[0]):
                v_max[0] = params_som.v_spike

        return v_min, v_max

    @staticmethod
    def make_lif(C_som, g_leak_som, E_rev_leak, input_mul=1.):
        """
        This function creates a DendriticParameters instance describing the
        parameters of a single-compartment, current-based LIF neuron.
        """

        A = np.array(((0., 0.), )) / C_som
        a_const = np.array((-g_leak_som, )) / C_som

        B = np.array(((input_mul, -input_mul), )) / C_som
        b_const = np.array((E_rev_leak * g_leak_som, )) / C_som

        C = np.array(((0., ), )) / C_som

        return DendriticParameters(A, a_const, B, b_const, C)

    @staticmethod
    def make_lif_cond(C_som=1e-9,
                      g_leak_som=50e-9,
                      E_rev_leak=-65e-3,
                      E_rev_exc=20e-3,
                      E_rev_inh=-75e-3,
                      input_mul=1.):
        """
        This function creates a DendriticParameters instance describing the
        parameters of a single-compartment, conductance-based LIF neuron.
        """

        A = np.array(((-input_mul, -input_mul), )) / C_som
        a_const = np.array((-g_leak_som, )) / C_som

        B = np.array(
            ((input_mul * E_rev_exc, input_mul * E_rev_inh), )) / C_som
        b_const = np.array((E_rev_leak * g_leak_som, )) / C_som

        C = np.array(((0., ), )) / C_som

        return DendriticParameters(A, a_const, B, b_const, C)

    @staticmethod
    def make_two_comp_lif(C_som=1e-9,
                          C_den=1e-9,
                          g_leak_som=50e-9,
                          g_leak_den=50e-9,
                          g_couple=50e-9,
                          E_rev_leak=-65e-3,
                          E_rev_exc=20e-3,
                          E_rev_inh=-75e-3,
                          input_mul=1.):
        """
        This function creates a DendriticParameters instance describing the
        parameters of a two-compartment LIF neuron.
        """

        Cs = np.array((C_som, C_den))

        A = np.array(((0., 0.), (-input_mul, -input_mul))) / Cs[:, None]
        a_const = np.array(
            (-(g_leak_som + g_couple), -(g_leak_den + g_couple))) / Cs

        B = np.array(
            ((0., 0.),
             (input_mul * E_rev_exc, input_mul * E_rev_inh))) / Cs[:, None]
        b_const = np.array(
            (E_rev_leak * g_leak_som, E_rev_leak * g_leak_den)) / Cs

        C = np.array(((0., g_couple), (g_couple, 0.))) / Cs[:, None]

        return DendriticParameters(A, a_const, B, b_const, C)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from nengo_bio.neurons import MultiChannelNeuronType

import nengo.builder
from nengo.builder.operator import Operator

class SimMultiChannelNeurons(Operator):
    def __init__(self, neurons, step_math, xs, output, tag=None):
        super().__init__(tag=tag)
        self.neurons = neurons
        self.step_math = step_math

        self.sets = [output]
        self.incs = []
        self.reads = [x for x in xs]
        self.updates = []

    @property
    def xs(self):
        return self.reads

    @property
    def output(self):
        return self.sets[0]

    @property
    def states(self):
        return self.sets[1:]

    def _descstr(self):
        return '%s, %s, %s' % (self.neurons, self.xs, self.output)

    def make_step(self, signals, dt, rng):
        xs = [signals[x] for x in self.xs]
        output = signals[self.output]
        def step_simneurons():
            self.step_math(output, *xs)
        return step_simneurons


@nengo.builder.Builder.register(MultiChannelNeuronType)
def build_neurons(model, neuron_type, neurons):
    # Compile the step_math function for this neuron type
    tuning = model.params[neurons.ensemble].tuning
    step_math = neuron_type.compile(model.dt, neurons.size_in, tuning)

    # Add the MultiChannelNeuronType simulator operator
    n_inputs = neuron_type.n_inputs
    xs = [model.sig[neurons]['in_{}'.format(i)] for i in range(n_inputs)]
    model.add_op(SimMultiChannelNeurons(neurons=neuron_type,
                                        step_math=step_math,
                                        xs=xs,
                                        output=model.sig[neurons]['out']))



#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import collections

from nengo_bio.common import steal_param, Excitatory, Inhibitory

import nengo.ensemble
import nengo.builder

class Ensemble(nengo.ensemble.Ensemble):
    """
    Wrapper class for creating a Dale's Principle enabled neuron population.
    All parameters are passed to the nengo Population object, except for the new
    p_exc and p_inh parameters. These parameters indicate the relative number of
    excitatory and inhibitory neurons.
    """

    @staticmethod
    def _coerce_p_exc_inh(p_exc, p_inh, ens=None):
        has_p_exc, has_p_inh = not p_exc is None, not p_inh is None
        if has_p_exc or has_p_inh:
            # Both p_exc and p_inh are given must sum to one
            if has_p_exc and has_p_inh:
                if abs(p_exc + p_inh - 1.0) > 1e-3:
                    raise ValueError(
                        "p_exc={} and p_inh={} do not add up to one for {}"
                        .format(p_exc, p_inh, ens)
                    )

            # At least p_exc is given, check range
            if has_p_exc:
                if p_exc < 0.0 or p_exc > 1.0:
                    raise ValueError(
                        "p_exc={} must be between 0.0 and 1.0 for {}"
                        .format(p_exc, ens))
                p_inh = 1.0 - p_exc
            # At least p_inh is given, check range
            elif has_p_inh:
                if p_inh < 0.0 or p_inh > 1.0:
                    raise ValueError(
                        "p_inh={} must be between 0.0 and 1.0 for {}"
                        .format(p_inh, ens))
                p_exc = 1.0 - p_inh
        return p_exc, p_inh

    def __init__(self, *args, **kw_args):
        """
        Forwards all parameters except for p_exc and p_inh to the parent
        Ensemble class.
        """

        self._p_exc, _ = Ensemble._coerce_p_exc_inh(
            steal_param('p_exc', kw_args), steal_param('p_inh', kw_args), self)

        super(Ensemble, self).__init__(*args, **kw_args)

    @property
    def p_exc(self):
        return self._p_exc

    @property
    def p_inh(self):
        if self._p_exc is None:
            return None
        else:
            return 1.0 - self._p_exc

    @p_exc.setter
    def p_exc(self, value):
        self._p_exc, _ = Ensemble._coerce_p_exc_inh(value, None, self)

    @p_inh.setter
    def p_inh(self, value):
        self._p_exc, _ = Ensemble._coerce_p_exc_inh(None, value, self)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

def steal_param(param_name, kw_args, default=None):
    if param_name in kw_args:
        value = kw_args[param_name]
        del kw_args[param_name]
        return value
    else:
        return default

class SynapseType:
    """
    The SynapseType class can be used to mark neurons as either excitatory or
    inhibitory.
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

Excitatory = SynapseType("excitatory")
Inhibitory = SynapseType("inhibitory")




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='nengo_bio',
    packages=find_packages(),
    package_data={
        'nengo_bio.internal':
        ['multi_compartment_lif.hpp']
    },
    version='0.2.0',
    author='Andreas Stöckel',
    author_email='astoecke@uwaterloo.ca',
    description='Dendritic Computation Primitives for Nengo',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/astoeckel/nengo-bio',
    license='GPLv3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ],
    install_requires=[
        "cvxopt>=1.2.2",
        "nengo>=2.8",
        "numpy>=1.16.3",
        "scipy>=1.2.0",
        "posix_ipc>=1.0.4",
    ],
)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from nengo_bio.internal.multi_compartment_lif_sim import make_simulator_class


def compile_simulator_python(params_som, params_den, dt=1e-3, ss=10):
    # Some handy aliases
    pS, pD = params_som, params_den

    # Compute v_min, v_max
    v_min, v_max = params_den.vEq_extreme(params_som)

    class PythonImpl:
        @staticmethod
        def run_step_from_memory(self, out, *xs):
            # Iterate over all neurons
            for i in range(self.n_neurons):
                # Fetch the input data
                x = np.array([xs[j][i] for j in range(self.n_inputs)])

                # Compute the A-matrix and the b-vector for this sample
                A = pD.C + np.diag(pD.a_const + pD.A @ x)
                b = pD.b_const + pD.B @ x

                # Access to the current state: membrane poential and
                # refractoriness
                S = self.state[i]

                # Write the initial output value
                out[i] = 0.

                # Advance the simulation for the given number of subsamples
                for s in range(ss):
                    # Compute the membrane potentials in dt / ss
                    S[:-1] += (A @ S[:-1] + b) * (dt / ss)

                    # Clamp the potentials
                    S[:-1] = np.clip(S[:-1], v_min, v_max)

                    # Handle refractoriness
                    if S[-1] > 0.0:
                        S[-1] -= dt / ss
                        S[0] = pS.v_spike if S[-1] > pS.tau_ref else pS.v_reset

                    # Handle spikes
                    if S[0] > pS.v_th and S[-1] <= 0.0:
                        S[-1] = pS.tau_ref + pS.tau_spike
                        S[0] = pS.v_spike if pS.tau_spike > 0 else pS.v_reset
                        out[i] = 1. / dt

        @staticmethod
        def run_single_with_constant_input(self, out, xs):
            for i in range(out.shape[0]):
                PythonImpl.run_step_from_memory(self, out[i:i + 1], *xs)

        @staticmethod
        def run_single_with_poisson_sources(self, out, sources):
            n_samples = out.size
            n_inputs = len(sources)

            # Initialize the individual random engines for the input channels,
            # pre-compute some filter constants
            rngs = [None] * n_inputs
            dist_exp, dist_gain = [None] * n_inputs, [None] * n_inputs
            filt, xs, offs, T = np.zeros((4, n_inputs, 1))
            for j, src in enumerate(sources):
                # Initialize the random engine for this input with the seed
                # specified by the user
                rngs[j] = np.random.RandomState(src.seed)

                # Compute the filter coefficient
                filt[j] = 1.0 - self.dt / src.tau

                # Setup the poisson and uniform distribution
                def mk_dists(j, src):
                    scale = 1.0 / (src.tau * src.rate)
                    dist_exp = lambda: rngs[j].exponential(1. / src.rate)
                    dist_gain = lambda: rngs[j].uniform(src.gain_min * scale, src.gain_max * scale)
                    return dist_exp, dist_gain

                dist_exp[j], dist_gain[j] = mk_dists(j, src)

                # Draw the first spike time
                T[j] = dist_exp[j]()

                # Setup the uniform gain distribution and initialize xs to the
                # average value
                xs[j] = 0.5 * (src.gain_min + src.gain_max)

                # Copy the offset
                offs[j] = src.offs

            # Implement the Poisson source and run the simulation
            for i in range(n_samples):
                # Apply the exponential filter
                xs *= filt

                # Simulate the Poisson spike source
                curT = i * self.dt
                for j in range(n_inputs):
                    while T[j] < curT:
                        # Feed a Delta pulse into the input
                        xs[j] += dist_gain[j]()

                        # Compute the next spike time
                        T[j] += dist_exp[j]()

                # Advance the simulation by one step
                PythonImpl.run_step_from_memory(self, out[i:i + 1],
                                                *(xs + offs))

        @staticmethod
        def run_single_with_gaussian_sources(self, out, sources):
            n_samples = out.size
            n_inputs = len(sources)

            # Initialize the individual random engines for the input channels,
            # pre-compute some filter constants
            rngs = [None] * n_inputs
            dist_norm = [None] * n_inputs
            filt, xs, offs = np.zeros((3, n_inputs, 1))
            for j, src in enumerate(sources):
                # Initialize the random engine for this input with the seed
                # specified by the user
                rngs[j] = np.random.RandomState(src.seed)

                # Compute the filter coefficient
                filt[j] = 1.0 - self.dt / src.tau

                # Setup the Gaussian distribution
                def mk_dists(j, src):
                    scale = self.dt / src.tau
                    dist_norm = lambda: rngs[j].normal(scale * src.mu, scale * src.sigma)
                    return dist_norm

                dist_norm[j] = mk_dists(j, src)

                # Setup the uniform gain distribution and initialize xs to the
                # average value
                xs[j] = src.mu

                # Copy the offset
                offs[j] = src.offs

            # Implement the Gaussian source and run the simulation
            for i in range(n_samples):
                for j in range(n_inputs):
                    xs[j] = xs[j] * filt[j] + dist_norm[j]()
                inp = np.maximum(0.0, xs + offs)
                PythonImpl.run_step_from_memory(self, out[i:i + 1], *inp)

    return make_simulator_class(PythonImpl, params_som, params_den, dt, ss)




#   nengo_bio -- Extensions to Nengo for more biological plausibility
#   Copyright (C) 2019  Andreas Stöckel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import ctypes
import numpy as np


class PoissonSource(ctypes.Structure):
    _fields_ = [("seed", ctypes.c_uint32), ("rate", ctypes.c_double),
                ("gain_min", ctypes.c_double), ("gain_max", ctypes.c_double),
                ("tau", ctypes.c_double), ("offs", ctypes.c_double)]


class GaussianSource(ctypes.Structure):
    _fields_ = [("seed", ctypes.c_uint32), ("mu", ctypes.c_double),
                ("sigma", ctypes.c_double), ("tau", ctypes.c_double),
                ("offs", ctypes.c_double)]


def make_simulator_class(impl, params_som_, params_den_, dt_, ss_):
    """
    The make_simulator_class creates a new Simulator class as used by both the
    Python and C++ simulator backend. It injects the given step_math function
    into a wrapper that performs additional type checking.

    Parameters
    ==========

    impl:
        Object providing the actual implementation of the simulator functions.
    params_som_:
        The somatic parameters that should be stored in the Simulator class.
    params_den_:
        The dendritic parameters that should be stored in the Simulator class.
    dt_:
        The simulation timestep.
    ss_:
        The number of subsamples per timestep used by the simulator.
    """

    def check(a, size):
        return (a.flags.c_contiguous and (a.dtype == np.float64)
                and a.size >= size)

    class Simulator:
        params_som = params_som_
        params_den = params_den_

        n_comp = params_den_.n_comp
        n_inputs = params_den_.n_inputs

        dt = dt_
        ss = ss_

        def __init__(self, n_neurons=1):
            # Copy the number of neurons
            self.n_neurons = n_neurons

            # Initialize the state matrix
            self.state = np.empty((n_neurons, self.n_comp + 1),
                                  order='C',
                                  dtype=np.float64)
            self.state[:, :-1] = params_som_.v_reset
            self.state[:, -1] = 0

        def run_step_from_memory(self, out, *xs):
            # Make sure the output array has the correct layout and length
            assert check(out, self.n_neurons)

            # Make sure the input arrays have the correct layout and length
            assert len(xs) == self.n_inputs
            for x in xs:
                check(x, self.n_neurons)

            # Call the actual step_math function
            impl.run_step_from_memory(self, out, *xs)

        def run_single_with_constant_input(self, out, xs):
            # Make sure n_neurons is one -- this function only makes sense when
            # simulating a single neuron
            assert self.n_neurons == 1

            # Make sure the output array is valid
            assert check(out, 0)

            # Make sure the input array is valid
            assert check(xs, self.n_inputs)

            impl.run_single_with_constant_input(self, out, xs)

        def run_single_with_poisson_sources(self, out, sources):
            # Make sure n_neurons is one -- this function only makes sense when
            # simulating a single neuron
            assert self.n_neurons == 1

            # Make sure the output array is valid
            assert check(out, 0)

            # Make sure the sources array is valid and copy the data over to a
            # C array
            assert (len(sources) == self.n_inputs)
            c_sources = (PoissonSource * self.n_inputs)()
            for i in range(self.n_inputs):
                c_sources[i].seed = sources[i].seed
                c_sources[i].rate = sources[i].rate
                c_sources[i].gain_min = sources[i].gain_min
                c_sources[i].gain_max = sources[i].gain_max
                c_sources[i].tau = sources[i].tau
                c_sources[i].offs = sources[i].offs

            impl.run_single_with_poisson_sources(self, out, c_sources)

        def run_single_with_gaussian_sources(self, out, sources):
            # Make sure n_neurons is one -- this function only makes sense when
            # simulating a single neuron
            assert self.n_neurons == 1

            # Make sure the output array is valid
            assert check(out, 0)

            # Make sure the sources array is valid and copy the data over to a
            # C array
            assert (len(sources) == self.n_inputs)
            c_sources = (GaussianSource * self.n_inputs)()
            for i in range(self.n_inputs):
                c_sources[i].seed = sources[i].seed
                c_sources[i].mu = sources[i].mu
                c_sources[i].sigma = sources[i].sigma
                c_sources[i].tau = sources[i].tau
                c_sources[i].offs = sources[i].offs

            impl.run_single_with_gaussian_sources(self, out, c_sources)

    return Simulator

